@@ -15,7 +15,7 @@
  *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
  *
  *  Copyright (C) 1999-2004 Olivier Fourdan (fourdan@xfce.org)
- *  Copyright (C) 2011 Peter de Ridder (peter@xfce.org)
+ *  Copyright (C) 2011-2012 Peter de Ridder (peter@xfce.org)
  *
  *  Portions based Thinice port by
  *                       Tim Gerla <timg@rrv.net>,
@@ -57,6 +57,9 @@
 #define GRIP_STYLE "grip-style"
 #define XFCE_GRIP_STYLE "-"XFCE_NAMESPACE"-"GRIP_STYLE
 
+#define BUTTON_DEFAULT_BORDER "button-default-border"
+#define XFCE_BUTTON_DEFAULT_BORDER "-"XFCE_NAMESPACE"-"BUTTON_DEFAULT_BORDER
+
 /* macros to make sure that things are sane ... */
 #define GE_CAIRO_INIT                               \
     cairo_set_line_width (cr, 1.0);                 \
@@ -65,86 +68,12 @@
 
 G_DEFINE_DYNAMIC_TYPE(XfceEngine, xfce_engine, GTK_TYPE_THEMING_ENGINE)
 
-/* Taken from raleigh theme engine */
-typedef enum
-{
-    CHECK_LIGHT,
-    CHECK_DARK,
-    CHECK_BASE,
-    CHECK_TEXT,
-    CHECK_CROSS,
-    CHECK_DASH,
-    RADIO_LIGHT,
-    RADIO_DARK,
-    RADIO_BASE,
-    RADIO_TEXT
-}
-Part;
-
-#define PART_SIZE 13
-
-static const guint32 check_light_bits[] = {
-    0x0000, 0x0000, 0x0800, 0x0800, 0x0800, 0x0800, 0x0800, 0x0800, 0x0800,
-    0x0800, 0x0800, 0x0ffc, 0x0000,
-};
-static const guint32 check_dark_bits[] = {
-    0x0000, 0x0ffe, 0x0002, 0x0002, 0x0002, 0x0002, 0x0002, 0x0002, 0x0002,
-    0x0002, 0x0002, 0x0002, 0x0000,
-};
-static const guint32 check_base_bits[] = {
-    0x0000, 0x0000, 0x07fc, 0x07fc, 0x07fc, 0x07fc, 0x07fc, 0x07fc, 0x07fc,
-    0x07fc, 0x07fc, 0x0000, 0x0000,
-};
-static const guint32 check_text_bits[] = {
-    0x0000, 0x0000, 0x1c00, 0x0f00, 0x0380, 0x01c0, 0x00e0, 0x0073, 0x003f,
-    0x003e, 0x001c, 0x0018, 0x0008
-};
-static const guint32 check_cross_bits[] = {
-    0x0000, 0x0000, 0x0000, 0x0300, 0x0380, 0x01d8, 0x00f8, 0x0078, 0x0038,
-    0x0018, 0x0000, 0x0000, 0x0000,
-};
-static const guint32 check_dash_bits[] = {
-    0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x03f8, 0x03f8, 0x03f8, 0x0000,
-    0x0000, 0x0000, 0x0000, 0x0000,
-};
-static const guint32 radio_light_bits[] = {
-    0x0000, 0x0000, 0x0000, 0x0400, 0x0800, 0x0800, 0x0800, 0x0800, 0x0800,
-    0x0400, 0x0208, 0x01f0, 0x0000,
-};
-static const guint32 radio_dark_bits[] = {
-    0x0000, 0x01f0, 0x0208, 0x0004, 0x0002, 0x0002, 0x0002, 0x0002, 0x0002,
-    0x0004, 0x0000, 0x0000, 0x0000,
-};
-static const guint32 radio_base_bits[] = {
-    0x0000, 0x0000, 0x01f0, 0x03f8, 0x07fc, 0x07fc, 0x07fc, 0x07fc, 0x07fc,
-    0x03f8, 0x01f0, 0x0000, 0x0000,
-};
-static const guint32 radio_text_bits[] = {
-    0x0000, 0x0000, 0x0000, 0x0000, 0x00e0, 0x01f0, 0x01f0, 0x01f0, 0x00e0,
-    0x0000, 0x0000, 0x0000, 0x0000,
-};
-
-static struct
-{
-    const guint32 *bits;
-    cairo_surface_t *bmap;
-}
-parts[] =
-{
-    { check_light_bits, NULL },
-    { check_dark_bits,  NULL },
-    { check_base_bits,  NULL },
-    { check_text_bits,  NULL },
-    { check_cross_bits, NULL },
-    { check_dash_bits,  NULL },
-    { radio_light_bits, NULL },
-    { radio_dark_bits,  NULL },
-    { radio_base_bits,  NULL },
-    { radio_text_bits,  NULL }
-};
+#define CHECK_MIN_SIZE 15
+#define CHECK_DRAW_SIZE 11
 
 /* internal functions */
 static void xfce_draw_grips(GtkThemingEngine * engine, cairo_t * cr, gdouble x, gdouble y, gdouble width, gdouble height, GtkOrientation orientation);
+static void xfce_draw_frame(GtkThemingEngine * engine, cairo_t * cr, gdouble x, gdouble y, gdouble width, gdouble height, GtkBorderStyle border_style);
 
 static void render_line(GtkThemingEngine * engine, cairo_t * cr, gdouble x1, gdouble y1, gdouble x2, gdouble y2);
 static void render_background(GtkThemingEngine * engine, cairo_t * cr, gdouble x, gdouble y, gdouble width, gdouble height);
@@ -377,6 +306,9 @@ static void render_line(GtkThemingEngine * engine, cairo_t * cr, gdouble x1, gdo
         x1 += 1 + thickness_dark - thickness_light;
         y_2 += 1;
 
+        y_1 = floor(y_1);
+        y_2 = floor(y_2);
+
         cairo_set_line_width (cr, thickness_dark);
         gdk_cairo_set_source_rgba(cr, &dark);
         cairo_move_to(cr, x1 - (thickness_dark / 2.0), y_1);
@@ -398,6 +330,9 @@ static void render_line(GtkThemingEngine * engine, cairo_t * cr, gdouble x1, gdo
         y_1 += 1 + thickness_dark - thickness_light;
         x2 += 1;
 
+        x1 = floor(x1);
+        x2 = floor(x2);
+
         cairo_set_line_width (cr, thickness_dark);
         gdk_cairo_set_source_rgba(cr, &dark);
         cairo_move_to(cr, x1, y_1 - (thickness_dark / 2.0));
@@ -444,7 +379,13 @@ static void render_background(GtkThemingEngine * engine, cairo_t * cr, gdouble x
 
     xt = MIN(xthick, width - 1);
     yt = MIN(ythick, height - 1);
-    if(smooth_edge && gtk_theming_engine_has_class(engine, GTK_STYLE_CLASS_PROGRESSBAR))
+    if (gtk_theming_engine_has_class(engine, GTK_STYLE_CLASS_NOTEBOOK))
+    {
+        xt = 0;
+        yt = 0;
+    }
+    else if ((smooth_edge && gtk_theming_engine_has_class(engine, GTK_STYLE_CLASS_PROGRESSBAR)) ||
+             (smooth_edge && gtk_theming_engine_has_class(engine, GTK_STYLE_CLASS_TROUGH)))
     {
         xt = 1;
         yt = 1;
@@ -452,10 +393,20 @@ static void render_background(GtkThemingEngine * engine, cairo_t * cr, gdouble x
     else
     {
         xt = MIN(xt, 2);
-        yt = MIN(yt, 2);
+        xt = MIN(xt, yt);
+        yt = xt;
     }
 
-    cairo_rectangle(cr, xt, yt, width - xt * 2, height - yt * 2);
+    /* The menubar only draws a bottom line */
+    if (gtk_theming_engine_has_class(engine, GTK_STYLE_CLASS_MENUBAR) &&
+        !gtk_theming_engine_has_class(engine, GTK_STYLE_CLASS_MENUITEM))
+    {
+        cairo_rectangle(cr, 0, 0, width, height - yt);
+    }
+    else
+    {
+        cairo_rectangle(cr, xt, yt, width - xt * 2, height - yt * 2);
+    }
 
     if(pattern)
     {
@@ -484,13 +435,58 @@ static void render_frame(GtkThemingEngine * engine, cairo_t * cr, gdouble x, gdo
     gint xthick, ythick;
     GtkStateFlags state;
     GtkBorderStyle border_style;
+    GtkBorder border;
+    GtkBorder *default_border;
+
+    state = gtk_theming_engine_get_state(engine);
+    gtk_theming_engine_get(engine, state, GTK_STYLE_PROPERTY_BORDER_STYLE, &border_style, NULL);
+
+    xthick = border.left;
+    ythick = border.top;
+
+    xt = MIN(xthick, width - 1);
+    yt = MIN(ythick, height - 1);
+
+    /* Spin buttons are a special case */
+    if (gtk_theming_engine_has_class(engine, GTK_STYLE_CLASS_SPINBUTTON) && gtk_theming_engine_has_class(engine, GTK_STYLE_CLASS_BUTTON))
+    {
+        /* Draw an outset border when hovering a spinner button */
+        if (!(state & GTK_STATE_FLAG_ACTIVE))
+            border_style = GTK_BORDER_STYLE_OUTSET;
+    }
+
+    /* Default buttons are a special case */
+    if (gtk_theming_engine_has_class(engine, GTK_STYLE_CLASS_BUTTON) && gtk_theming_engine_has_class(engine, GTK_STYLE_CLASS_DEFAULT))
+    {
+        /* Draw an inset border around the default border */
+        gtk_theming_engine_get(engine, state, XFCE_BUTTON_DEFAULT_BORDER, &default_border, NULL);
+
+        if (default_border &&
+            (default_border->left > xt) && (default_border->right > xt) &&
+            (default_border->top > yt) && (default_border->bottom > yt))
+        {
+            xfce_draw_frame(engine, cr, x - default_border->left, y - default_border->top,
+                    width + default_border->left + default_border->right, height + default_border->top + default_border->bottom,
+                    GTK_BORDER_STYLE_INSET);
+        }
+
+        gtk_border_free(default_border);
+    }
+
+    xfce_draw_frame(engine, cr, x, y, width, height, border_style);
+}
+
+static void xfce_draw_frame(GtkThemingEngine * engine, cairo_t * cr, gdouble x, gdouble y, gdouble width, gdouble height, GtkBorderStyle border_style)
+{
+    gint xt, yt;
+    gint xthick, ythick;
+    GtkStateFlags state;
     GdkRGBA dark, light, mid, bg;
     GdkRGBA black = {0.0, 0.0, 0.0, 1.0}; /* black */
     gboolean smooth_edge;
     GtkBorder border;
 
     state = gtk_theming_engine_get_state(engine);
-    gtk_theming_engine_get(engine, state, GTK_STYLE_PROPERTY_BORDER_STYLE, &border_style, NULL);
 
     if (border_style == GTK_BORDER_STYLE_NONE)
         return;
@@ -503,21 +499,21 @@ static void render_frame(GtkThemingEngine * engine, cairo_t * cr, gdouble x, gdo
     xthick = border.left;
     ythick = border.top;
 
-    /* Spin buttons are a special case */
-    if (gtk_theming_engine_has_class(engine, GTK_STYLE_CLASS_SPINBUTTON) && gtk_theming_engine_has_class(engine, GTK_STYLE_CLASS_BUTTON))
-    {
-        /* Draw an outset border when hovering a spinner button */
-        if (!(state & GTK_STATE_FLAG_ACTIVE))
-            border_style = GTK_BORDER_STYLE_OUTSET;
-    }
-
     xt = MIN(xthick, width - 1);
     yt = MIN(ythick, height - 1);
 
     switch (border_style)
     {
+#if GTK_CHECK_VERSION(3,4,0)
+        case GTK_BORDER_STYLE_HIDDEN:
+#endif
         case GTK_BORDER_STYLE_NONE:
             break;
+#if GTK_CHECK_VERSION(3,4,0)
+        case GTK_BORDER_STYLE_DOTTED:
+        case GTK_BORDER_STYLE_DASHED:
+        case GTK_BORDER_STYLE_DOUBLE:
+#endif
         case GTK_BORDER_STYLE_SOLID:
             gdk_cairo_set_source_rgba(cr, &dark);
             if ((xt > 1) && (yt > 1))
@@ -531,14 +527,96 @@ static void render_frame(GtkThemingEngine * engine, cairo_t * cr, gdouble x, gdo
             }
             cairo_stroke(cr);
             break;
+#if GTK_CHECK_VERSION(3,4,0)
+	case GTK_BORDER_STYLE_GROOVE:
+            color_dark2light(&dark, &light);
+            if ((xt > 1) && (yt > 1))
+            {
+                gdk_cairo_set_source_rgba(cr, &light);
+                cairo_move_to(cr, x + 0.5, y + height - 0.5);
+                cairo_line_to(cr, x + width - 0.5, y + height - 0.5);
+                cairo_line_to(cr, x + width - 0.5, y + 0.5);
+                cairo_stroke(cr);
+
+                gdk_cairo_set_source_rgba(cr, &dark);
+                cairo_move_to(cr, x + width - 1.5, y + 0.5);
+                cairo_line_to(cr, x + 0.5, y + 0.5);
+                cairo_line_to(cr, x + 0.5, y + height - 1.5);
+                cairo_stroke(cr);
+
+                gdk_cairo_set_source_rgba(cr, &light);
+                cairo_move_to(cr, x + width - 1.5, y + 1.5);
+                cairo_line_to(cr, x + 1.5, y + 1.5);
+                cairo_line_to(cr, x + 1.5, y + height - 1.5);
+                cairo_stroke(cr);
+
+                gdk_cairo_set_source_rgba(cr, &dark);
+                cairo_move_to(cr, x + 1.5, y + height - 1.5);
+                cairo_line_to(cr, x + width - 1.5, y + height - 1.5);
+                cairo_line_to(cr, x + width - 1.5, y + 1.5);
+                cairo_stroke(cr);
+            }
+            else if ((xt > 0) && (yt > 0))
+            {
+                gdk_cairo_set_source_rgba(cr, &dark);
+                cairo_rectangle(cr, x + 0.5, y + 0.5, width - 1, height - 1);
+                cairo_stroke(cr);
+            }
+            break;
+	case GTK_BORDER_STYLE_RIDGE:
+            color_dark2light(&dark, &light);
+            if ((xt > 1) && (yt > 1))
+            {
+                gdk_cairo_set_source_rgba(cr, &dark);
+                cairo_move_to(cr, x + 0.5, y + height - 0.5);
+                cairo_line_to(cr, x + width - 0.5, y + height - 0.5);
+                cairo_line_to(cr, x + width - 0.5, y + 0.5);
+                cairo_stroke(cr);
+
+                gdk_cairo_set_source_rgba(cr, &light);
+                cairo_move_to(cr, x + width - 1.5, y + 0.5);
+                cairo_line_to(cr, x + 0.5, y + 0.5);
+                cairo_line_to(cr, x + 0.5, y + height - 1.5);
+                cairo_stroke(cr);
+
+                gdk_cairo_set_source_rgba(cr, &dark);
+                cairo_move_to(cr, x + width - 1.5, y + 1.5);
+                cairo_line_to(cr, x + 1.5, y + 1.5);
+                cairo_line_to(cr, x + 1.5, y + height - 1.5);
+                cairo_stroke(cr);
+
+                gdk_cairo_set_source_rgba(cr, &light);
+                cairo_move_to(cr, x + 1.5, y + height - 1.5);
+                cairo_line_to(cr, x + width - 1.5, y + height - 1.5);
+                cairo_line_to(cr, x + width - 1.5, y + 1.5);
+                cairo_stroke(cr);
+            }
+            else if ((xt > 0) && (yt > 0))
+            {
+                gdk_cairo_set_source_rgba(cr, &light);
+                cairo_rectangle(cr, x + 0.5, y + 0.5, width - 1, height - 1);
+                cairo_stroke(cr);
+            }
+            break;
+#endif
         case GTK_BORDER_STYLE_INSET:
             gtk_theming_engine_get(engine, state, XFCE_SMOOTH_EDGE, &smooth_edge, NULL);
             color_dark2light_mid(&dark, &light, &mid);
             if (smooth_edge)
             {
-                if (gtk_theming_engine_has_class(engine, GTK_STYLE_CLASS_TROUGH))
+                if (gtk_theming_engine_has_class(engine, GTK_STYLE_CLASS_TROUGH) && !gtk_theming_engine_has_class(engine, GTK_STYLE_CLASS_SCALE))
                 {
-                    /* Do nothing */
+                    gtk_theming_engine_get_background_color(engine, state, &bg);
+                    gdk_cairo_set_source_rgba(cr, &bg);
+                    cairo_rectangle(cr, x + 0.5, y + 0.5, width - 1, height - 1);
+                    cairo_stroke(cr);
+
+                    gdk_cairo_set_source_rgba(cr, &mid);
+                    cairo_rectangle(cr, x, y, 1, 1);
+                    cairo_rectangle(cr, x + width - 1, y, 1, 1);
+                    cairo_rectangle(cr, x, y + height - 1, 1, 1);
+                    cairo_rectangle(cr, x + width - 1, y + height - 1, 1, 1);
+                    cairo_fill(cr);
                 }
                 else if ((xt > 1) && (yt > 1))
                 {
@@ -600,7 +678,6 @@ static void render_frame(GtkThemingEngine * engine, cairo_t * cr, gdouble x, gdo
             {
                 if (gtk_theming_engine_has_class(engine, GTK_STYLE_CLASS_TROUGH))
                 {
-                    gtk_theming_engine_get_border_color(engine, GTK_STATE_FLAG_ACTIVE, &dark);
                     gdk_cairo_set_source_rgba(cr, &dark);
                     cairo_rectangle(cr, x + 0.5, y + 0.5, width - 1, height - 1);
                     cairo_stroke(cr);
@@ -662,9 +739,6 @@ static void render_frame(GtkThemingEngine * engine, cairo_t * cr, gdouble x, gdo
                     cairo_rectangle(cr, x + 0.5, y + 0.5, width - 1, height - 1);
                     cairo_stroke(cr);
 
-                    gtk_theming_engine_get_border_color(engine, GTK_STATE_FLAG_NORMAL, &dark);
-                    color_dark2light_mid(&dark, &light, &mid);
-
                     gdk_cairo_set_source_rgba(cr, &mid);
                     cairo_rectangle(cr, x, y, 1, 1);
                     cairo_rectangle(cr, x + width - 1, y, 1, 1);
@@ -681,7 +755,8 @@ static void render_frame(GtkThemingEngine * engine, cairo_t * cr, gdouble x, gdo
                         cairo_stroke(cr);
                     }
                 }
-                else if (gtk_theming_engine_has_class(engine, GTK_STYLE_CLASS_MENUBAR))
+                else if (gtk_theming_engine_has_class(engine, GTK_STYLE_CLASS_MENUBAR) &&
+                         !gtk_theming_engine_has_class(engine, GTK_STYLE_CLASS_MENUITEM))
                 {
                     if ((xt > 1) && (yt > 1))
                     {
@@ -736,7 +811,7 @@ static void render_frame(GtkThemingEngine * engine, cairo_t * cr, gdouble x, gdo
                         cairo_stroke(cr);
                     }
                 }
-                else if (gtk_theming_engine_has_class(engine, GTK_STYLE_CLASS_SLIDER))
+                else if (gtk_theming_engine_has_class(engine, GTK_STYLE_CLASS_SCROLLBAR) || gtk_theming_engine_has_class(engine, GTK_STYLE_CLASS_SLIDER) || gtk_theming_engine_has_class(engine, GTK_STYLE_CLASS_SCALE))
                 {
                     if ((xt > 1) && (yt > 1))
                     {
@@ -884,7 +959,8 @@ static void render_frame(GtkThemingEngine * engine, cairo_t * cr, gdouble x, gdo
                 {
                     /* Do nothing */
                 }
-                else if (gtk_theming_engine_has_class(engine, GTK_STYLE_CLASS_MENUBAR))
+                else if (gtk_theming_engine_has_class(engine, GTK_STYLE_CLASS_MENUBAR) &&
+                         !gtk_theming_engine_has_class(engine, GTK_STYLE_CLASS_MENUITEM))
                 {
                     if ((xt > 1) && (yt > 1))
                     {
@@ -984,47 +1060,117 @@ static void render_frame(GtkThemingEngine * engine, cairo_t * cr, gdouble x, gdo
     }
 }
 
-static cairo_surface_t *get_part_bmap (Part part)
+static void draw_dash(cairo_t * cr, const GdkRGBA * c, gdouble x, gdouble y, guint size)
 {
-    if (!parts[part].bmap)
+    guint w, b;
+
+    b = (size + 7) / 10;
+
+    w = size / 4;
+    if ((w % 2) != (size % 2))
     {
-        parts[part].bmap = cairo_image_surface_create_for_data((guchar*)parts[part].bits, CAIRO_FORMAT_A1, PART_SIZE, PART_SIZE, sizeof(guint32));
+        w += 1;
     }
-    return parts[part].bmap;
-}
 
-static void draw_part(cairo_t * cr, const GdkRGBA * c, gdouble x, gdouble y, Part part)
-{
     gdk_cairo_set_source_rgba(cr, c);
 
-    cairo_mask_surface(cr, get_part_bmap (part), x, y);
+    cairo_set_line_width (cr, w);
+    cairo_set_line_cap(cr, CAIRO_LINE_CAP_BUTT);
+
+    cairo_move_to (cr, x + b, y + size / 2.0);
+    cairo_line_to (cr, x + size - b, y + size / 2.0);
+
+    cairo_stroke(cr);
 }
 
 static void render_check(GtkThemingEngine * engine, cairo_t * cr, gdouble x, gdouble y, gdouble width, gdouble height)
 {
     GtkStateFlags state;
     GdkRGBA bg, border, fg;
+    guint size;
+    guint w, b;
 
-    x -= floor((1 + PART_SIZE - width) / 2);
-    y -= floor((1 + PART_SIZE - height) / 2);
+    /* Make sure it doesn't get to small */
+    if (width < CHECK_MIN_SIZE)
+        width = CHECK_DRAW_SIZE;
+    else
+    {
+        width -= CHECK_MIN_SIZE - CHECK_DRAW_SIZE;
+        x += (CHECK_MIN_SIZE - CHECK_DRAW_SIZE) / 2;
+    }
+    if (height < CHECK_MIN_SIZE)
+        height = CHECK_DRAW_SIZE;
+    else
+    {
+        height -= CHECK_MIN_SIZE - CHECK_DRAW_SIZE;
+        y += (CHECK_MIN_SIZE - CHECK_DRAW_SIZE) / 2;
+    }
+
+    /* Make it square */
+    if (width > height)
+    {
+        x += width - height;
+        size = height;
+    }
+    else
+    {
+        y += height - width;
+        size = width;
+    }
 
     state = gtk_theming_engine_get_state(engine);
     gtk_theming_engine_get_background_color(engine, state, &bg);
     gtk_theming_engine_get_border_color(engine, state, &border);
     gtk_theming_engine_get_color(engine, state, &fg);
 
+    GE_CAIRO_INIT;
+
+    cairo_set_antialias (cr, CAIRO_ANTIALIAS_NONE);
+
+    cairo_rectangle (cr, x + 0.5, y + 0.5, size - 1, size - 1);
+
     if (!gtk_theming_engine_has_class(engine, GTK_STYLE_CLASS_MENUITEM))
-        draw_part(cr, &bg, x, y, CHECK_BASE);
-    draw_part(cr, &border, x, y, CHECK_LIGHT);
-    draw_part(cr, &border, x, y, CHECK_DARK);
+    {
+        /* Draw the background */
+        gdk_cairo_set_source_rgba(cr, &bg);
+        cairo_fill_preserve(cr);
+    }
+
+    /* Draw the border */
+    gdk_cairo_set_source_rgba(cr, &border);
+    cairo_stroke(cr);
+
+    x += 1;
+    y += 1;
+    size -= 2;
 
     if (state & GTK_STATE_FLAG_INCONSISTENT)
     {
-        draw_part(cr, &fg, x, y, CHECK_DASH);
+        draw_dash(cr, &fg, x, y, size);
     }
     else if (state & GTK_STATE_FLAG_ACTIVE)
     {
-        draw_part(cr, &fg, x, y, CHECK_CROSS);
+        b = (size + 7) / 10;
+        w = ((size + 4 - b) / 6);
+
+        /* Draw the check */
+        gdk_cairo_set_source_rgba(cr, &fg);
+
+        cairo_move_to (cr, x + b, y + floor(size / 2 - 1.5));
+
+        cairo_line_to (cr, x + b, y + size - b);
+        cairo_line_to (cr, x + b + w, y + size - b);
+
+        cairo_line_to (cr, x + size - b, y + b + w);
+        cairo_line_to (cr, x + size - b, y + b);
+        cairo_line_to (cr, x + size - b + 1 - w, y + b);
+
+        cairo_line_to (cr, x + b + w, y + size - b + 1 - 2 * w);
+
+        cairo_line_to (cr, x + b + w, y + floor(size / 2 - 1.5));
+
+        cairo_close_path (cr);
+        cairo_fill(cr);
     }
 }
 
@@ -1032,27 +1178,75 @@ static void render_option(GtkThemingEngine * engine, cairo_t * cr, gdouble x, gd
 {
     GtkStateFlags state;
     GdkRGBA bg, border, fg;
+    guint size;
 
-    x -= (1 + PART_SIZE - width) / 2;
-    y -= (1 + PART_SIZE - height) / 2;
+    /* Make sure it doesn't get to small */
+    if (width < CHECK_MIN_SIZE)
+        width = CHECK_DRAW_SIZE;
+    else
+    {
+        width -= CHECK_MIN_SIZE - CHECK_DRAW_SIZE;
+        x += (CHECK_MIN_SIZE - CHECK_DRAW_SIZE) / 2;
+    }
+    if (height < CHECK_MIN_SIZE)
+        height = CHECK_DRAW_SIZE;
+    else
+    {
+        height -= CHECK_MIN_SIZE - CHECK_DRAW_SIZE;
+        y += (CHECK_MIN_SIZE - CHECK_DRAW_SIZE) / 2;
+    }
+
+    /* Make it square */
+    if (width > height)
+    {
+        x += width - height;
+        size = height;
+    }
+    else
+    {
+        y += height - width;
+        size = width;
+    }
 
     state = gtk_theming_engine_get_state(engine);
     gtk_theming_engine_get_background_color(engine, state, &bg);
     gtk_theming_engine_get_border_color(engine, state, &border);
     gtk_theming_engine_get_color(engine, state, &fg);
 
+    GE_CAIRO_INIT;
+
+    cairo_set_antialias (cr, CAIRO_ANTIALIAS_NONE);
+
+    /* This is required since the render layout leaves a stale cairo point */
+    cairo_new_sub_path (cr);
+    cairo_arc (cr, x + (size / 2.0), y + (size / 2.0), (size - 1) / 2.0, 0, 2 * M_PI);
+
     if (!gtk_theming_engine_has_class(engine, GTK_STYLE_CLASS_MENUITEM))
-        draw_part(cr, &bg, x, y, RADIO_BASE);
-    draw_part(cr, &border, x, y, RADIO_LIGHT);
-    draw_part(cr, &border, x, y, RADIO_DARK);
+    {
+        /* Draw the background */
+        gdk_cairo_set_source_rgba(cr, &bg);
+        cairo_fill_preserve(cr);
+    }
+
+    /* Draw the border */
+    gdk_cairo_set_source_rgba(cr, &border);
+    cairo_stroke(cr);
+
+    x += 1;
+    y += 1;
+    size -= 2;
 
     if (state & GTK_STATE_FLAG_INCONSISTENT)
     {
-        draw_part(cr, &fg, x, y, CHECK_DASH);
+        draw_dash(cr, &fg, x, y, size);
     }
     else if (state & GTK_STATE_FLAG_ACTIVE)
     {
-        draw_part(cr, &fg, x, y, RADIO_TEXT);
+        /* Draw the dot */
+        gdk_cairo_set_source_rgba(cr, &fg);
+
+        cairo_arc (cr, x + (size / 2.0), y + (size / 2.0), (size / 2.0) - ((size + 2) / 5), 0, 2 * M_PI);
+        cairo_fill(cr);
     }
 }
 
@@ -1350,6 +1544,8 @@ static void render_frame_gap(GtkThemingEngine * engine, cairo_t * cr, gdouble x,
             ew = 2;
             eh = gap_e - gap_s;
             break;
+        default:
+            return;
     }
 
     cairo_save (cr);
@@ -1361,7 +1557,7 @@ static void render_frame_gap(GtkThemingEngine * engine, cairo_t * cr, gdouble x,
     cairo_rectangle (cr, x0, ey + eh, x1 - x0, y_1 - (ey + eh));
     cairo_clip (cr);
 
-    render_frame (engine, cr, x, y, width, height);
+    xfce_draw_frame (engine, cr, x, y, width, height, border_style);
 
     cairo_restore (cr);
 }
@@ -1381,7 +1577,7 @@ static void render_extension(GtkThemingEngine * engine, cairo_t * cr, gdouble x,
     state = gtk_theming_engine_get_state(engine);
     gtk_theming_engine_get_border(engine, state, &border);
 
-    GTK_THEMING_ENGINE_GET_CLASS(engine)->render_background(engine, cr, x - border.left + 1, y - border.top + 1, width + border.left + border.right - 2, height + border.top + border.bottom - 2);
+    GTK_THEMING_ENGINE_GET_CLASS(engine)->render_background(engine, cr, x, y, width, height);
 
     gtk_theming_engine_get(engine, state, XFCE_SMOOTH_EDGE, &smooth_edge, NULL);
 
@@ -1574,6 +1770,13 @@ static void xfce_engine_class_init(XfceEngineClass * klass)
     gtk_theming_engine_register_property(XFCE_NAMESPACE, NULL,
             g_param_spec_enum(GRIP_STYLE, "Grip style", "Grip style",
                 XFCE_TYPE_GRIP_STYLE, XFCE_GRIP_ROUGH, 0));
+
+    /* Compatibility properties */
+    gtk_theming_engine_register_property(XFCE_NAMESPACE, NULL,
+            g_param_spec_boxed (BUTTON_DEFAULT_BORDER,
+                "Default Spacing",
+                "Extra space to add for GTK_CAN_DEFAULT buttons",
+                GTK_TYPE_BORDER, 0));
 }
 
 static void xfce_engine_class_finalize(XfceEngineClass * klass)
